// ctrlr914 - ATtiny84 timing using timer1 interrupt - PWM tone output

// Started: Feb 27, 2015  G. D. (Joe) Young <jyoung@islandnet.com>
//
// Revised: Apr 26/17 - From ctrlr14H, int service modified for fast PWM
//          Apr 28/17 - tone generator working
//          Apr 30/17 - tone generator macros in ctrlr14.h
//          May  1/17 - 1 sec earlier ptt output
//          May 11/17 - fine ptt delay adjust, fast flashes in delay (from ctrlr14)
//          May 24/17 - ctrlr14.h revised, SSS definition (from ctrlr14)
//          May 26/17 - add adjOverlap macro
//          May 30/17 - eeprom endint, cleanups
//          Jun  4/17 - keep pwm pin active toneon flag for on/off control
//          Jun  6/17 - tone control options TONE_PTT, TONE_CONT
//
//
// ctrlr914 will include all the compile-time options that have been tested
// in other versions of codppmx.... In addition, the modification to incorporate fox ID
// on the FoxOr continuous mode is included, which requires more modes be available
// ctrlr914 uses the 14-pin processor so that inputs are available to select a start
// delay. During the start delay, the controller only sends a short message every
// 5 minutes and then the normal timing for the selected mode begains after the start
// delay has expired.
//
// The designation 914 is adopted to suggest the derivation from the earlier 2m
// controller which used PWM tone generation. However, here the timer must also keep
// its function to provide accurate 1 second ticks for transmission interval timing.
//
//
// To permit re-programming the callsign without otherwise modifying the controller
// code, the fox ID (MOE..MO5) and the call are stored in processor's eeprom at 
// location 0, in the form of morse codewords (see codebytes.txt).

// An example eeprom entry:
// >>> dump eeprom 
// 0000  07 0f 08 00 0c 02 00 11  02 38 18 12 0d 32 38 ff  |.... ....8.. 28.|
// which corresponds to the text: MOS de VE7BFK/7

// The codeword storage format is used to avoid storing the table of ASCII-to-morse
// An arduino utility, text2code, can be used to produce an intel-hex file for loading
// into eeprom.

// There are 8 modes of operation:
//  fmode  operation       typical assignment (see ctrlr14.h for actual assignments)
//   FOS  FoxOr slow            0
//   SPS  Sprint slow           1
//   BEA  continuous MO         2
//   NOR  normal 1min/4min      3
//   NU1  not used              4
//   NU2  not used              5
//   SPF  Sprint fast           6
//   FOF  FoxOr fast            7


#include <avr/io.h>
#include <avr/interrupt.h>
#include <EEPROM.h>
#include "sin_tab.h"

// image of first few eeprom locations. Compiler produces .eep file, but arduino
// programmer command doesn't load it.
byte eedata[] __attribute__ ((section(".eeprom")))=
     {0x07,0x0f,0x04,0x00,0x11,0x05,0x38,0x0f,0x07,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
      0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
      0x05,0x00,0x0a,0x00,0x0f,0x00,0x14,0x00,0xff,0xff,0xff,0xff,0xff,0xff,0x05,0x00,
      0xda,0x02,0x72,0x1C,0x57,0x02};

// ***** compile-time options
#define INDEXMODENOR         // calculate start time from fox number in normal mode

#define TPOLARITY_POS      // Tx output active, positive true
//#define TPOLARITY_NEG      // Tx output active, negative true
//#define TPOLARITY_OD       // Tx output open-drain negative true
//#define LPOLARITY_POS      // LED output active, positive true
#define LPOLARITY_OD        // LED output open-drain negative true
#define PTPOLARITY_POS      // PTT line active pos true
//#define PTPOLARITY_NEG      // PTT line active neg true
//#define PTPOLARITY_OD       // PTT line open-drain negative true

#define CALL_INTERVAL 1800    // time interval between sending callsign

//#define TEST 1              // output waveform on named arduino pin

#define NRMODEBITS 2        // may be 3, 2, or 1 - see ctrlr14.h
char modselectpins[ ] = { 0x08, 0x07, 0x06 };  // pins to use for mode selection

#define NRDELAYBITS 2
char dlyselectpins[ ] = { 0x02, 0x03, 0x01 };  // pins to use for delay selection

#define DLYINT 5      // interval between burst Tx during delayed start (multiple of 5)
//#define DLYSENDCALL   // include call sending during delay bursts

#define ADJ_OVERLAP 1500  // adjust ON/OFF times by length of MOx. 1000 no overlap

#define TONE_PIN 6    // specify tone on pin 5 or 6, freq by incr value in addr 0x32
//#define TONE_PTT      // tone on/off with ptt line
//#define TONE_CONT     // tone on at setup, runs continuously

const byte ledPin = 5;
const byte txPin = 4;
const byte ainPin = A0;
const byte pttPin = 1;
const byte vbattPin = A0;    // need to figure out how to deal with duplicates

// bring in definitions from specified compile-time options above
#include "ctrlr14.h"


byte fmode, thisfox;
word onTime, offTime, ledtime, calltime;
boolean code = false;              // true during code sending
boolean sendm, sendo, msent, sendid, sendsp, cend, ledon;
boolean sendlsp, sendwsp, eltOn;   // loop control of morse output
unsigned long timer;
unsigned long dotLen = 221;        // dot length in millisec (1200/wpm)1.8432
unsigned long cdotLen = 55;  // dot length while sending call sign
unsigned long savedotLen;


// new functions and variables for callsign insertion
void setupNext( byte );       // functions to setup pointers to code bytes
void setupCall( );
byte setGetBit( );    // code byte parsing, now returns fence bit position 7..0
bool getBit( );
byte codeChar, nextelt, nextBit, bitCnt, nextChar, thisLength;
byte callend;
bool endCode, callsend;

// new functions and variables for delayed start feature
void saveSetup( );
void restoreSetup( );
byte inputPinToBin( byte, char * );
int  dlyTime;         // time interval for delay in minutes
int  dlyLastint;      // duration of last interval during which burst Tx happens
word saveOn, saveOff; // for setup settings saved during delay
word saveLoopOvfl;    // save start times
byte dlyIndex;        // pin selection of delay interval stored in EEPROM
bool dlyStart;        // flag to indicate start is delayed
bool endint = false;  // indicates the end part of delay interval

// battery test functions and variables
void checkVbatt( );
int lowBattCount;     // A/D count for low battery voltage threshold
int battime;          // minute counter for battery test interval continuous modes
bool lowBatt = false; // flag set when input voltage goes below lowBattCount

// PTT (oscillator control) line timing
int pttADJ;           // interval in msec*1.8432 between ptt and first code
bool pttadj = false;  // timing to adjust ptt start is underway

// interrupt service variables
volatile word ovflcnt = 0;
volatile byte seccnt = 0;
volatile bool testtog, toneon;
volatile word minutes, samplecnt, accumulator, increment;
word loopovfl, fivemin, onemin, savePTTtime, onesec, pttTimer;
unsigned long offctr, pttmsec;
byte highbyte;


ISR( TIM1_OVF_vect ) {

if( toneon ) {
  accumulator += increment;
  highbyte = accumulator >> 8;
//  OCR1A = sinTab[highbyte];
  toneSample( );
} else {
  toneStop( );
}

  samplecnt += 1;
  if( samplecnt >= 7200 ) {
    samplecnt = 0;
    
    ovflcnt += 1;
    seccnt += 1;
    if( seccnt == 60 ) {
      seccnt = 0;
      minutes += 1;          // add minute count for delayed start timing
    }
  #ifdef TEST
    testtog = !testtog;
    digitalWrite( TEST, testtog );
  #endif
  } // test count at sample rate
}

void setupTimer1( ) {
  setTone( );
  TCCR1B = ( bit(WGM13) | bit(WGM12) | bit(CS10) );  // clk div 1
  ICR1 = 255;
  OCR1A = 127;
  accumulator = 0;
  EEPROM.get( 0x32, increment );  // 7282==>800 Hz; 9102==>1000 Hz
  TIMSK1 = bit(TOIE1);
}

void setup( ) {

// analog input mode selection
//  int temp = analogRead( ainPin );      // get operating mode
//  fmode = temp>>7;                      // Mar 3/17 - 8 modes, keep top 3 bits
//  if( NRMODEBITS < 3 ) fmode = fmode>>1;
//  if( NRMODEBITS == 1 ) fmode = fmode>>1;

// digital input mode selection
  fmode = inputPinToBin( NRMODEBITS, modselectpins );
  if( fmode==NU1 ) fmode = FOS;         // default to continuous for not used
//  if( fmode==NU2 ) fmode = FOF;   // now used for SSS

// check inputs for delayed start
  dlyTime = 0;
  dlyStart = false;
  
// here read pins and compose index for delay lookup, lookup delay
  dlyIndex = inputPinToBin( NRDELAYBITS, dlyselectpins );
  EEPROM.get( 2*dlyIndex+0x1E, dlyTime );   // lookup delay, first one at 0x20
  if( dlyTime <= 0 ) {
    dlyIndex = 0;
  } // if bad time read from EE

  if( dlyIndex != 0 ) {
    dlyTime += (int)DLYINT;           // count down starts at zero
    dlyStart = true;
    EEPROM.get( 0x2e, dlyLastint );
    if( dlyLastint <= 0 ) dlyLastint = dlyTime;
  }

// read low battery threshold count from EEPROM, disable if invalid
  EEPROM.get( 0x30, lowBattCount );
  if( lowBattCount < 0 || lowBattCount > 1020 ) lowBattCount = 0;

// read ptt advance time from EEPROM, default if invalid
  int pttTMP;
  EEPROM.get( 0x34, pttTMP );
  if( pttTMP < 0 || pttTMP > 1820 ) pttTMP = 1000;
  pttADJ = 1840 - pttTMP;

  onTime = 12;
  offTime = 48;
  if( fmode == SPF || fmode == FOF ) dotLen -= 64;        // fast sprint, FoxOr
  savedotLen = dotLen;
  if( fmode == NOR ) {
    onTime = 60;
    offTime = 240;
  } // if normal timing

  setLEDpin( );
  setTxpin( );
  setPTTpin( );
  eltOn = false;
  cend = false;
  ledon = true;

  codeChar = EEPROM.read( 2 );  //determine fox number from EE code character
  thisfox = setGetBit( );
  callend = 4;                  //locate end of callsign
  while( EEPROM.read( callend ) != 0xff ) callend++;

  thisLength = 3;
  if( fmode == BEA || fmode == SSS ) {
    thisfox = 0;              // no fox id for continuous
    thisLength = 2;
    ledtime = ovflcnt + 30;   // 30 sec for led 
  }
  if( fmode==FOF || fmode==FOS ) ledtime = ovflcnt + 30;  //led at start of FoxOr modes

// Setup timers
  if( fmode == SPF || fmode == SPS ) {
    loopovfl = ovflcnt+1+(thisfox-1)*onTime;  //indexed start for sprint
  } else {
    loopovfl = ovflcnt + 1;    // no indexing for normal
  } // if sprint modes

#ifdef INDEXMODENOR
// Also index start time from fox number if in normal mode
  if( fmode == NOR ) {
    loopovfl = ovflcnt+1+(thisfox-1)*onTime;  //indexed start for normal
  } // if mode NOR to be indexed start
#endif

  adjOverlap( );    // reduce on time by length MOx to prevent overlap 

  callsend = true;

  saveSetup( );
  pttTimer = loopovfl -1;

  if( dlyStart ) {
    callsend = false;
    dotLen = cdotLen;
    onTime = 1;                 // for single MOx during no call delay
  #ifdef DLYSENDCALL
    onTime = 5;
  #endif
    offTime = 60*DLYINT - onTime;
  }

 #ifdef TEST
  pinMode( TEST, OUTPUT );      // enable test timing signal on named pin
 #endif

  noInterrupts( );    // attempt to sync millis( ) with timer1
  setupTimer1( );
  interrupts( );
  toneCONTON( );      // tone on continuously

} // setup( )


void loop( ) {

  if( lowBatt && minutes == fivemin ) {
    fivemin = minutes+DLYINT;
    loopovfl = ovflcnt + saveLoopOvfl;
    pttTimer = loopovfl-1;
  }

  if( code && lowBatt && loopovfl==ovflcnt && (fmode==BEA || fmode==FOF || fmode==FOS || fmode==SSS) ) {
    loopovfl += offTime;
    fivemin = minutes+DLYINT;
    cend = true;    // delay end until id complete
  }

  if( (fmode==BEA || fmode==FOF || fmode==FOS || fmode==SSS) && minutes-battime >= 1 ) {
    battime = minutes;
    if( !lowBatt ) checkVbatt( );
    if( lowBatt ) cend = true;
  }

  if( dlyStart && dlyTime > 0 && minutes == fivemin ) {
    fivemin = minutes+DLYINT;
#ifdef DLYSENDCALL
    callsend = true;
#endif
    dlyTime -= (int)DLYINT;
    if( dlyTime <= dlyLastint ) endint = true;
    if( dlyTime <= 0 ) {
      restoreSetup( );
      dlyStart = false;
      endint = false;
      pttTimer = loopovfl-1;
    }
    if( endint ) {
      dotLen = cdotLen;
      loopovfl = ovflcnt + saveLoopOvfl;
      pttTimer = loopovfl-1;
    } 
  }

// turn ON led at each minute during the delayed start interval
  if( dlyStart && (ovflcnt - onesec) >= 1 ) {
    LEDpinON( );
    onesec = ovflcnt;
    offctr = millis( );
  }

// turn OFF led after flash duration
  if( dlyStart && (millis( ) - offctr) >= 500 ){
    LEDpinOFF( );
  }


  if( !code && pttTimer == ovflcnt && (!dlyStart || endint) ) { // turn on PTT line early
    pttmsec = millis( );
    pttadj = true;              // start timing adjustment timer
    pttTimer += onTime+offTime;
  }

  if( pttadj && (millis( ) - pttmsec) >= pttADJ ) {
    PTTpinON( );
    tonePTTON( );
    pttadj = false;
  }

  if( !code && loopovfl == ovflcnt && (!dlyStart || endint ) ) {  // start sending
    loopovfl += onTime;
    nextChar = 0;
    setupNext( nextChar );      // get first letter of fox code
    code = true;
    cend = false;
  }

// end sending
  if( code && loopovfl==ovflcnt && (fmode==SPS || fmode==SPF || fmode==NOR || dlyStart ) ) {
    loopovfl += offTime;
    cend = true;    // delay end until id complete
  }

  if( (fmode==BEA || fmode==FOF || fmode==FOS || fmode==SSS) && ledtime == ovflcnt ) ledon = false;
  
  if( !dlyStart && (ovflcnt-calltime) > CALL_INTERVAL ) {
    calltime = ovflcnt;
    callsend = true;
  }
  
  if( code && !eltOn ) {
    timer = millis( ) + dotLen;            // setup for dot length
    if( getBit( ) ) timer += 2*dotLen;      // sending a dash
    TxpinON( );
    toneON( );
    if( ledon ) LEDpinON( );
    eltOn = true;
  } // if not sending, start first element

  if( code && eltOn && !sendsp && ( timer < millis( ) ) ) {
    TxpinOFF( );
    LEDpinOFF( );
    toneOFF( );
    timer += dotLen;    // timing space following element
    sendsp = true;
    if( endCode ) {
      endCode = false;
      sendlsp = true;
      timer += 2*dotLen;
      nextChar++;
      if( (nextChar == thisLength) || (nextChar == callend) ) {
        dotLen = savedotLen;
        sendwsp = true;
        timer += 3*dotLen;
        nextChar = 0;          // reset pointer to characters
      } // if last character of this group was just sent
      setupNext( nextChar );
      if( codeChar == 0 ) {
        sendwsp = true;
        timer += 3*dotLen;
        nextChar++;
        if( nextChar == thisLength ) nextChar = 0;
        setupNext( nextChar );    // if space charcter to be sent
      }
    } // extend space to letter space or word space
  } // if element on time is up

  if( code && sendsp && ( timer < millis( ) ) ) {
    eltOn = false;
    sendsp = false;
    if( sendlsp ) {
      sendlsp = false;
    } // if letter space just ended
    if( sendwsp ) {
      sendwsp = false;
      if( cend ) {
        code = false;  // turn off sending
        ledon = false;   // and led after first burst
        PTTpinOFF( );
        tonePTTOFF( );
        if( !lowBatt ) checkVbatt( );     //once in lowBatt, stay there
      } // if last word space, end sending
      if( callsend ) {
        callsend = false;
        setupCall( );
      } // if time to send a callsign
    } // if word space ended, setup to begin again
  } // if sending space 
  
} // loop( )


// support function implementations
byte ssschars[ ] = { 0x08, 0x00, 0x00 };
void setupNext( byte nextChar ) {
  codeChar = EEPROM.read( nextChar );
  if( fmode == SSS && nextChar < 4 ) {
    codeChar = ssschars[ nextChar ];
  }
  setGetBit( );
} // setup to first code byte (usually for M)

void setupCall( ) {
  nextChar =  4;    // start of callsign in EE
  codeChar = EEPROM.read( nextChar );
  dotLen = cdotLen;
  if( codeChar == 0xff ) {              // allow for suppressing call
    nextChar = 0;
    codeChar = EEPROM.read( nextChar );
    dotLen = savedotLen;
  }
  setGetBit( );
} // setup to first code byte of callsign ID

byte setGetBit( ) {
  bitCnt = 7;
  nextBit = 0b10000000;
  while( (bitCnt > 0) && ((nextBit & codeChar) == 0) ) {
    nextBit = nextBit>>1;
    bitCnt--;
  } // while leading zeros before fence bit
  return bitCnt;
} // setup the code word parsing

bool getBit( ) {
  nextBit = nextBit>>1;
  bitCnt--;
  if( bitCnt == 0 ) {
    endCode = true;
  } // if last bit has just been selected
  return( (nextBit & codeChar) == nextBit );
} // get next code element to send

// support for delayed start
void saveSetup( ) {
  saveOn = onTime;
  saveOff = offTime;
  saveLoopOvfl = loopovfl;    // indexed fox start times
  savePTTtime = loopovfl-1;
}

void restoreSetup( ) {
  onTime = saveOn;
  offTime = saveOff;
  loopovfl = ovflcnt + saveLoopOvfl;
  calltime = ovflcnt+2;
  dotLen = savedotLen;
}

//read input pins into binary byte MSB first
byte inputPinToBin( byte nr, char* pins ) {
  byte bin = 0xff;
  for( byte ix=0; ix<nr; ix++ ) {
    pinMode( pins[ix], INPUT_PULLUP );
    bool bit = digitalRead( pins[ix] );
    if( bit ) {
      bin |= 0x01;
    } else {
      bin &= 0xfe;
    }
    if( ix < nr-1 ) bin = bin<<1;
  }
  return ~bin;
}


void checkVbatt( ) {
  int vbatt = analogRead( vbattPin );
  if( vbatt < lowBattCount ) {
    lowBatt = true;
    if( fmode == NOR ) {
      fivemin = minutes + DLYINT - 1 - saveLoopOvfl/60;
    } else {
      fivemin = minutes + 1;
    }
    onTime = 6;
    offTime = 60*DLYINT - onTime;
  } else {
    lowBatt = false;
  }
}

